import isNil from '@tinkoff/utils/is/nil';
import compose from '@tinkoff/utils/function/compose';
import { mergeOptions, createTinkoffRequest, HttpClientAdapter } from '@tramvai/tinkoff-request-http-client-adapter';
import { fillHeaderIp, fillHeaders } from '../utils/fillHeaders.es.js';
import { createUserAgent } from './createUserAgent.es.js';

const environmentDependentOptions = typeof window === 'undefined'
    ? {
        defaultTimeout: 2000,
    }
    : {
        defaultTimeout: 30000,
    };
const httpClientFactory = ({ logger, envManager, appInfo, requestManager, headersList, createCache, makeRequestRegistry, agent, querySerializer, disableCircuitBreaker = false, defaultOptions, defaultInterceptors, commandLineExecutionContext, }) => {
    return (options) => {
        var _a;
        if (!options.name) {
            throw Error(`You need to pass a unique field "name" for the HTTP client instance`);
        }
        const forceDisableCache = envManager.get('HTTP_CLIENT_CACHE_DISABLED');
        const forceDisabledCircuitBreaker = envManager.get('HTTP_CLIENT_CIRCUIT_BREAKER_DISABLED');
        const interceptors = { interceptors: defaultInterceptors || [] };
        const adapterOptions = mergeOptions(mergeOptions({
            logger,
            agent,
            querySerializer,
            method: 'GET',
            // don't create function here to prevent leak `commandLineExecutionContext` (which refer to Child DI) into the new function closure
            createCache: createCache
                ? createCache.bind(null, 'memory')
                : undefined,
            modifyRequest: compose(fillHeaderIp({ requestManager }), fillHeaders({ requestManager, headersList })),
            circuitBreakerOptions: {
                failureThreshold: 75,
                minimumFailureCount: 10,
            },
            // TODO: remove any after [resolving](https://github.com/southpolesteve/node-abort-controller/issues/31)
            signal: (_a = commandLineExecutionContext === null || commandLineExecutionContext === void 0 ? void 0 : commandLineExecutionContext()) === null || _a === void 0 ? void 0 : _a.abortSignal,
            ...environmentDependentOptions,
        }, defaultOptions ? mergeOptions(defaultOptions, interceptors) : interceptors), options);
        // по умолчанию, на сервере, библиотека https://github.com/node-fetch/node-fetch
        // отправляет заголовок "User-Agent" вида "node-fetch".
        // для улучшения логов сервисов, в которые делают запросы tramvai приложения,
        // заменяем "User-Agent" на кастомный, содержащий название и версию приложения
        if (typeof window === 'undefined') {
            adapterOptions.headers = {
                'User-Agent': createUserAgent({ appInfo, envManager }),
                ...adapterOptions.headers,
            };
        }
        if (!isNil(forceDisableCache)) {
            adapterOptions.disableCache = !!forceDisableCache;
        }
        if (disableCircuitBreaker) {
            adapterOptions.enableCircuitBreaker = false;
        }
        // environment variable in priority over disableCircuitBreaker
        if (!isNil(forceDisabledCircuitBreaker)) {
            adapterOptions.enableCircuitBreaker = !forceDisabledCircuitBreaker;
        }
        // cache @tinkoff/request instance for performance reason
        if (!makeRequestRegistry.has(adapterOptions.name)) {
            const tRequestOptions = { ...adapterOptions };
            // prevent memory leak, because COMMAND_LINE_EXECUTION_CONTEXT_TOKEN has Request scope and linked to ChildContainer
            delete tRequestOptions.signal;
            makeRequestRegistry.set(adapterOptions.name, createTinkoffRequest(tRequestOptions));
        }
        const makeRequest = makeRequestRegistry.get(adapterOptions.name);
        const httpClientAdapter = new HttpClientAdapter({
            options: adapterOptions,
            makeRequest,
        });
        return httpClientAdapter;
    };
};

export { httpClientFactory };
