'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

var fetch = require('node-fetch');
var testHelpers = require('@tramvai/test-helpers');
var testMocks = require('@tramvai/test-mocks');
var dippy = require('@tinkoff/dippy');
var core = require('@tramvai/core');
var tokensHttpClient = require('@tramvai/tokens-http-client');
var tokensCommon = require('@tramvai/tokens-common');
var isNil = require('@tinkoff/utils/is/nil');
var compose = require('@tinkoff/utils/function/compose');
var tinkoffRequestHttpClientAdapter = require('@tramvai/tinkoff-request-http-client-adapter');
var identity = require('@tinkoff/utils/function/identity');
var flatten = require('@tinkoff/utils/array/flatten');
var pick = require('@tinkoff/utils/object/pick');
var tokensServer = require('@tramvai/tokens-server');
var find = require('@tinkoff/utils/array/find');
var httpClient = require('@tramvai/http-client');
var papi = require('@tramvai/papi');
var tokensServerPrivate = require('@tramvai/tokens-server-private');

function _interopDefaultLegacy (e) { return e && typeof e === 'object' && 'default' in e ? e : { 'default': e }; }

var fetch__default = /*#__PURE__*/_interopDefaultLegacy(fetch);
var isNil__default = /*#__PURE__*/_interopDefaultLegacy(isNil);
var compose__default = /*#__PURE__*/_interopDefaultLegacy(compose);
var identity__default = /*#__PURE__*/_interopDefaultLegacy(identity);
var flatten__default = /*#__PURE__*/_interopDefaultLegacy(flatten);
var pick__default = /*#__PURE__*/_interopDefaultLegacy(pick);
var find__default = /*#__PURE__*/_interopDefaultLegacy(find);

const fillHeaderIp = ({ requestManager, }) => {
    if (!requestManager) {
        return identity__default["default"];
    }
    return (params) => {
        return {
            ...params,
            headers: {
                ...params.headers,
                'X-real-ip': requestManager.getClientIp(),
            },
        };
    };
};
const fillHeaders = ({ requestManager, headersList, }) => {
    if (!requestManager) {
        return identity__default["default"];
    }
    const headerNames = flatten__default["default"](headersList !== null && headersList !== void 0 ? headersList : []);
    return (params) => {
        return {
            ...params,
            headers: {
                ...params.headers,
                ...pick__default["default"](headerNames, requestManager.getHeaders()),
            },
        };
    };
};

/** Compares real path with pattern path
 * @example comparePathWithPattern('/example/mock/12345', '/example/:type/:id') => true
 */
const comparePathWithPattern = (path, pattern) => new RegExp(`^${pattern.replace(/:(\w+)/g, '([^\\/]+)')}$`).test(path);

const getPathParams = (path, pattern) => { var _a, _b; return (_b = (_a = path.match(new RegExp(`^${pattern.replace(/:(\w+)/g, '(?<$1>[^\\/]+)')}$`))) === null || _a === void 0 ? void 0 : _a.groups) !== null && _b !== void 0 ? _b : {}; };

const createUserAgent = ({ appInfo, envManager, }) => {
    const { appName } = appInfo;
    const appVersion = envManager.get('APP_VERSION');
    const appRelease = envManager.get('APP_RELEASE');
    const userAgent = [
        'tramvai',
        appName,
        appVersion && `version ${appVersion}`,
        appRelease && `release ${appRelease}`,
    ]
        .filter(Boolean)
        .join(' ');
    return userAgent;
};

const environmentDependentOptions = typeof window === 'undefined'
    ? {
        defaultTimeout: 2000,
    }
    : {
        defaultTimeout: 30000,
    };
const httpClientFactory = ({ logger, envManager, appInfo, requestManager, headersList, createCache, makeRequestRegistry, agent, querySerializer, disableCircuitBreaker = false, defaultOptions, defaultInterceptors, commandLineExecutionContext, }) => {
    return (options) => {
        var _a;
        if (!options.name) {
            throw Error(`You need to pass a unique field "name" for the HTTP client instance`);
        }
        const forceDisableCache = envManager.get('HTTP_CLIENT_CACHE_DISABLED');
        const forceDisabledCircuitBreaker = envManager.get('HTTP_CLIENT_CIRCUIT_BREAKER_DISABLED');
        const interceptors = { interceptors: defaultInterceptors || [] };
        const adapterOptions = tinkoffRequestHttpClientAdapter.mergeOptions(tinkoffRequestHttpClientAdapter.mergeOptions({
            logger,
            agent,
            querySerializer,
            method: 'GET',
            // don't create function here to prevent leak `commandLineExecutionContext` (which refer to Child DI) into the new function closure
            createCache: createCache
                ? createCache.bind(null, 'memory')
                : undefined,
            modifyRequest: compose__default["default"](fillHeaderIp({ requestManager }), fillHeaders({ requestManager, headersList })),
            circuitBreakerOptions: {
                failureThreshold: 75,
                minimumFailureCount: 10,
            },
            // TODO: remove any after [resolving](https://github.com/southpolesteve/node-abort-controller/issues/31)
            signal: (_a = commandLineExecutionContext === null || commandLineExecutionContext === void 0 ? void 0 : commandLineExecutionContext()) === null || _a === void 0 ? void 0 : _a.abortSignal,
            ...environmentDependentOptions,
        }, defaultOptions ? tinkoffRequestHttpClientAdapter.mergeOptions(defaultOptions, interceptors) : interceptors), options);
        // по умолчанию, на сервере, библиотека https://github.com/node-fetch/node-fetch
        // отправляет заголовок "User-Agent" вида "node-fetch".
        // для улучшения логов сервисов, в которые делают запросы tramvai приложения,
        // заменяем "User-Agent" на кастомный, содержащий название и версию приложения
        if (typeof window === 'undefined') {
            adapterOptions.headers = {
                'User-Agent': createUserAgent({ appInfo, envManager }),
                ...adapterOptions.headers,
            };
        }
        if (!isNil__default["default"](forceDisableCache)) {
            adapterOptions.disableCache = !!forceDisableCache;
        }
        if (disableCircuitBreaker) {
            adapterOptions.enableCircuitBreaker = false;
        }
        // environment variable in priority over disableCircuitBreaker
        if (!isNil__default["default"](forceDisabledCircuitBreaker)) {
            adapterOptions.enableCircuitBreaker = !forceDisabledCircuitBreaker;
        }
        // cache @tinkoff/request instance for performance reason
        if (!makeRequestRegistry.has(adapterOptions.name)) {
            const tRequestOptions = { ...adapterOptions };
            // prevent memory leak, because COMMAND_LINE_EXECUTION_CONTEXT_TOKEN has Request scope and linked to ChildContainer
            delete tRequestOptions.signal;
            makeRequestRegistry.set(adapterOptions.name, tinkoffRequestHttpClientAdapter.createTinkoffRequest(tRequestOptions));
        }
        const makeRequest = makeRequestRegistry.get(adapterOptions.name);
        const httpClientAdapter = new tinkoffRequestHttpClientAdapter.HttpClientAdapter({
            options: adapterOptions,
            makeRequest,
        });
        return httpClientAdapter;
    };
};

class PapiService extends httpClient.BaseHttpClient {
    constructor({ papi, di }) {
        super();
        this.papi = flatten__default["default"](papi || []);
        this.di = di;
    }
    async request({ path, query, body, headers, }) {
        var _a;
        const pathWithLeadingSlash = (path === null || path === void 0 ? void 0 : path.startsWith('/')) ? path : `/${path}`;
        const papiRoute = find__default["default"]((papi$1) => comparePathWithPattern(pathWithLeadingSlash, papi.getPapiParameters(papi$1).path), (_a = this.papi) !== null && _a !== void 0 ? _a : []);
        if (!papiRoute) {
            throw new Error(`papi handler '${path}' not found`);
        }
        const req = {
            headers: { ...headers, host: 'localhost' },
            cookies: {},
            query,
            body,
            params: getPathParams(pathWithLeadingSlash, papi.getPapiParameters(papiRoute).path),
        };
        const res = {};
        const childDi = dippy.createChildContainer(this.di, [
            {
                provide: tokensServerPrivate.FASTIFY_REQUEST,
                useValue: req,
            },
            {
                provide: tokensServerPrivate.FASTIFY_RESPONSE,
                useValue: res,
            },
        ]);
        const papiExecutor = childDi.get(tokensServerPrivate.PAPI_EXECUTOR);
        const payload = await papiExecutor(papiRoute);
        return { payload, status: 200, headers: {} };
    }
}

const PapiClientModule = /* @__PURE__ */ core.Module({
    providers: [
        core.provide({
            provide: tokensHttpClient.PAPI_SERVICE,
            scope: core.Scope.SINGLETON,
            useClass: PapiService,
            deps: {
                di: core.DI_TOKEN,
                papi: { token: tokensServer.SERVER_MODULE_PAPI_PUBLIC_ROUTE, optional: true },
            },
        }),
    ],
})(class PapiClientModule {
});

const createCacheToken = dippy.createToken('httpClient createCache');
const HttpClientModule = /* @__PURE__ */ core.Module({
    imports: [PapiClientModule],
    providers: [
        core.provide({
            provide: tokensHttpClient.HTTP_CLIENT_FACTORY,
            useFactory: httpClientFactory,
            deps: {
                logger: tokensCommon.LOGGER_TOKEN,
                envManager: tokensCommon.ENV_MANAGER_TOKEN,
                appInfo: core.APP_INFO_TOKEN,
                createCache: createCacheToken,
                makeRequestRegistry: 'makeRequestRegistry',
                requestManager: {
                    token: tokensCommon.REQUEST_MANAGER_TOKEN,
                    optional: true,
                },
                headersList: {
                    token: tokensHttpClient.API_CLIENT_PASS_HEADERS,
                    optional: true,
                },
                agent: {
                    token: tokensHttpClient.HTTP_CLIENT_AGENT,
                    optional: true,
                },
                disableCircuitBreaker: {
                    token: tokensHttpClient.DISABLE_CIRCUIT_BREAKER,
                    optional: true,
                },
                defaultOptions: {
                    token: tokensHttpClient.DEFAULT_HTTP_CLIENT_FACTORY_OPTIONS,
                    optional: true,
                },
                defaultInterceptors: {
                    token: tokensHttpClient.DEFAULT_HTTP_CLIENT_INTERCEPTORS,
                    optional: true,
                },
                commandLineExecutionContext: {
                    token: tokensCommon.COMMAND_LINE_EXECUTION_CONTEXT_TOKEN,
                    optional: true,
                },
            },
        }),
        core.provide({
            provide: tokensHttpClient.HTTP_CLIENT,
            useFactory: ({ factory }) => {
                return factory({
                    name: 'http-client',
                    enableCircuitBreaker: false,
                });
            },
            deps: {
                factory: tokensHttpClient.HTTP_CLIENT_FACTORY,
            },
        }),
        core.provide({
            provide: tokensCommon.ENV_USED_TOKEN,
            useValue: [
                { key: 'HTTP_CLIENT_CACHE_DISABLED', optional: true, dehydrate: false },
                { key: 'HTTP_CLIENT_CIRCUIT_BREAKER_DISABLED', optional: true, dehydrate: false },
            ],
        }),
        /**
         * хранилище для экземпляров @tinkoff/request
         *
         * требуется хранить экземпляры в единственном виде,
         * т.к. многие плагины @tinkoff/request после инициализации имеют состояние
         * (cache, circuit breaker), и не будут корректно работать на сервере,
         * если создавать новые экземпляры на Scope.REQUEST
         */
        core.provide({
            provide: 'makeRequestRegistry',
            scope: core.Scope.SINGLETON,
            useFactory: () => new Map(),
        }),
        /**
         * `CREATE_CACHE_TOKEN` имеет проверку, если токен используется провайдером,
         * который имеет Scope.SINGLETON, то инстанс кэша сохраняется в общее хранилище,
         * и доступен для очистки через `/papi/clear-cache`.
         * Scope.REQUEST игнорируется, т.к. это верная утечка памяти,
         * инстансов кэши было бы неограниченное количество.
         *
         * HTTP клиенты создаются со Scope.REQUEST, но инстансы @tinkoff/request
         * (и соответственно кэшей) создаются только один раз, благодаря `makeRequestRegistry`.
         * это гарантирует отсутствие утечек памяти, поэтому мы обходим проверку
         * на Scope.SINGLETON c помощью обертки `createCacheToken`
         */
        core.provide({
            provide: createCacheToken,
            scope: core.Scope.SINGLETON,
            useFactory: ({ createCache }) => {
                return createCache;
            },
            deps: {
                createCache: tokensCommon.CREATE_CACHE_TOKEN,
            },
        }),
        core.provide({
            provide: tokensHttpClient.API_CLIENT_PASS_HEADERS,
            useValue: ['x-request-id'],
        }),
    ],
})(class HttpClientModule {
});

jest.mock('node-fetch');
const { Response } = jest.requireActual('node-fetch');
const testApi = (options) => {
    const caches = [];
    const { modules = [], providers = [], env } = options;
    const { di } = testHelpers.getDiWrapper({
        di: options.di,
        modules: [
            testMocks.CommonTestModule.forRoot({
                env,
                onCacheCreated: (cache) => {
                    caches.push(cache);
                },
            }),
            HttpClientModule,
            ...modules,
        ],
        providers: [...providers],
    });
    const fetchMock = fetch__default["default"];
    const clearCaches = () => {
        caches.forEach((cache) => cache.clear());
    };
    return {
        di,
        fetchMock,
        mockJsonResponse: async (body, init = {}) => {
            clearCaches();
            const { headers = {} } = init;
            fetchMock.mockImplementation(() => Promise.resolve(new Response(JSON.stringify(body), {
                status: 200,
                ...init,
                headers: {
                    'content-type': 'application/json',
                    ...headers,
                },
            })));
        },
        clearCaches,
    };
};

exports.testApi = testApi;
